// -*- Mode: C++; indent-tabs-mode: nil; c-basic-offset: 4; -*-
//
// OpenVRML
//
// Copyright (C) 1998  Chris Morley
// Copyright (C) 2003  Braden McDaniel
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

namespace openvrml {

    class node_interface {
    public:
        enum type_id {
            invalid_type_id,
            eventin_id,
            eventout_id,
            exposedfield_id,
            field_id
        };

        type_id type;
        field_value::type_id field_type;
        std::string id;

        node_interface(type_id type,
                       field_value::type_id field_type,
                       const std::string & id);
    };

    std::ostream & operator<<(std::ostream & out,
                                             node_interface::type_id type);
    std::istream & operator>>(std::istream & in,
                                             node_interface::type_id & type);

    inline bool operator==(const node_interface & lhs,
                           const node_interface & rhs)
        throw ()
    {
        return lhs.type == rhs.type
            && lhs.field_type == rhs.field_type
            && lhs.id == rhs.id;
    }

    inline bool operator!=(const node_interface & lhs,
                           const node_interface & rhs)
        throw ()
    {
        return !(lhs == rhs);
    }

    class node_type;

    class unsupported_interface : public std::runtime_error {
    public:
        explicit unsupported_interface(const std::string & message);
        unsupported_interface(const node_type & type,
                              const std::string & interface_id);
        unsupported_interface(const node_type & type,
                              node_interface::type_id interface_type,
                              const std::string & interface_id);
        virtual ~unsupported_interface() throw ();
    };


    class node_interface_set {
        struct id_less :
                std::binary_function<node_interface, node_interface, bool> {
            bool operator()(const node_interface & lhs,
                            const node_interface & rhs) const
            {
                return lhs.id < rhs.id;
            }
        };

        std::set<node_interface, id_less> interfaces;

    public:
        typedef std::set<node_interface, id_less>::const_iterator
            const_iterator;

        void add(const node_interface & interface)
            throw (std::invalid_argument, std::bad_alloc);
        const_iterator begin() const throw ();
        const_iterator end() const throw ();
        const_iterator find(const std::string & id) const throw ();
    };

    inline node_interface_set::const_iterator node_interface_set::begin() const
        throw ()
    {
        return this->interfaces.begin();
    }

    inline node_interface_set::const_iterator node_interface_set::end() const
        throw ()
    {
        return this->interfaces.end();
    }


/**
 * @class unsupported_interface
 *
 * @brief Exception to indicate that a node interface is not supported.
 *
 * This exception is thrown by node::field and node::process_event to
 * indicate that the node doesn't support the interface through which the
 * caller is trying to modify the node. It is also thrown by
 * node_class::create_type if the class object doesn't support an interface
 * specified in the node_interface_set given to that method.
 */

/**
 * @brief Construct.
 *
 * @param message   An informative error message.
 */
unsupported_interface::unsupported_interface(const std::string & message):
    std::runtime_error(message)
{}

/**
 * @brief Construct.
 *
 * @param type          the node type.
 * @param interface_id  the name of the interface that is not available.
 */
unsupported_interface::unsupported_interface(const node_type & type,
                                             const std::string & interface_id):
    std::runtime_error(type.id + " has no interface \"" + interface_id + '"')
{}

namespace {
    const std::string to_string(const node_interface::type_id interface_type)
    {
        std::ostringstream out;
        out << interface_type;
        return out.str();
    }
}

/**
 * @brief Construct.
 *
 * @param type              the node type.
 * @param interface_type    the type of the interface that is not available.
 * @param interface_id      the name of the interface that is not available.
 */
unsupported_interface::unsupported_interface(
    const node_type & type,
    const node_interface::type_id interface_type,
    const std::string & interface_id):
    std::runtime_error(type.id + " has no " + to_string(interface_type) + " \""
                       + interface_id + '"')
{}

/**
 * @brief Destroy.
 */
unsupported_interface::~unsupported_interface() throw ()
{}


/**
 * @class node_interface
 *
 * @brief Type information for an interface of a node.
 */

/**
 * @enum node_interface::type_id
 *
 * @brief Identify the type of interface.
 */

/**
 * @var node_interface::type_id node_interface::invalid_type_id
 *
 * @brief Used to indicate an invalid or unknown interface type.
 */

/**
 * @var node_interface::type_id node_interface::eventin_id
 *
 * @brief An eventIn.
 */

/**
 * @var node_interface::type_id node_interface::eventout_id
 *
 * @brief An eventOut.
 */

/**
 * @var node_interface::type_id node_interface::exposedfield_id
 *
 * @brief An exposedField.
 */

/**
 * @var node_interface::type_id node_interface::field_id
 *
 * @brief A field.
 */

namespace {
    const char * const node_interface_type_id_[] = {
        "<invalid interface type>",
        "eventIn",
        "eventOut",
        "exposedField",
        "field"
    };
}

/**
 * @brief Stream inserter.
 *
 * @relates node_interface
 *
 * If @p type is node_interface::invalidType, @c failbit is set on @p out.
 *
 * @param out   an output stream.
 * @param type  a node interface type.
 *
 * @return @p out.
 */
std::ostream & operator<<(std::ostream & out,
                          const node_interface::type_id type)
{
    if (type == node_interface::invalid_type_id) {
        out.setstate(std::ios_base::failbit);
    } else {
        out << node_interface_type_id_[type];
    }
    return out;
}

/**
 * @brief Stream extractor.
 *
 * @relates node_interface
 *
 * @param in    an input stream.
 * @param type  a node interface type.
 *
 * @return @p in.
 */
std::istream & operator>>(std::istream & in, node_interface::type_id & type)
{
    using std::find;
    using std::string;

    string interface_type_id;
    in >> interface_type_id;

    static const char * const * const begin =
            node_interface_type_id_ + node_interface::eventin_id;
    static const char * const * const end =
            node_interface_type_id_ + node_interface::field_id + 1;
    const char * const * const pos = find(begin, end, interface_type_id);
    if (pos != end) {
        type = node_interface::type_id(pos - begin);
    } else {
        in.setstate(std::ios_base::failbit);
    }
    return in;
}


/**
 * @var node_interface::type
 *
 * @brief The interface type.
 */

/**
 * @var node_interface::field_type
 *
 * @brief The field data type handled by this interface.
 */

/**
 * @var node_interface::id
 *
 * @brief The name of this interface.
 */

/**
 * @brief Constructor.
 *
 * @param type       the type of interface.
 * @param field_type the field data type handled by the interface.
 * @param id         the name of the interface.
 */
node_interface::node_interface(const type_id type,
                               const field_value::type_id field_type,
                               const std::string & id):
    type(type),
    field_type(field_type),
    id(id)
{}

/**
 * @fn bool operator==(const node_interface & lhs, const node_interface & rhs) throw ()
 *
 * @relates openvrml::node_interface
 *
 * @brief Compare for equality.
 *
 * @param lhs   a node_interface.
 * @param rhs   a node_interface.
 *
 * @return @c true if the two node_interfaces are equal, @c false otherwise.
 */

/**
 * @fn bool operator!=(const node_interface & lhs, const node_interface & rhs) throw ()
 *
 * @relates node_interface
 *
 * @brief Compare for openvrml::inequality.
 *
 * @param lhs   a node_interface.
 * @param rhs   a node_interface.
 *
 * @return @c true if the two node_interfaces are equal, @c false otherwise.
 */

/**
 * @class node_interface_set
 *
 * @brief A group of unique @link node_interface node_interfaces@endlink.
 *
 * node_interface_sets are used to construct new
 * @link node_type node_types@endlink. node_type objects also expose their
 * interfaces as a node_interface_set. The interfaces in a node_interface_set
 * are guaranteed to be unique and non-conflicting.
 */



/**
 * @internal
 *
 * @struct node_interface_set::id_less
 *
 * @brief A functor for ordering @link node_interface node_interfaces@endlink
 *        in the set.
 *
 * Compares the @a id's of two @link node_interface node_interfaces@endlink
 * using @c operator<.
 */

/**
 * @fn bool node_interface_set::id_less::operator()(const node_interface & lhs, const node_interface & rhs) const
 *
 * @brief Simple lexicographic comparison of the @a id's.
 *
 * @return @c true if @p lhs.id is lexicographically less than @p rhs.id,
 *      @c false otherwise.
 */

/**
 * @internal
 *
 * @var std::set<node_interface, node_interface_set::id_less> node_interface_set::interfaces
 *
 * @brief Set of @link node_interface node_interfaces@endlink.
 */

/**
 * @typedef node_interface_set::const_iterator
 *
 * @brief An STL const_iterator.
 */

/**
 * @brief Add an interface.
 *
 * @param interface a node_interface.
 *
 * @exception std::invalid argument if @p nodeInterface conflicts with an interface
 *                              already in the node_interface_set.
 * @exception std::bad_alloc        if memory allocation fails.
 */
void node_interface_set::add(const node_interface & interface)
    throw (std::invalid_argument, std::bad_alloc)
{
    if (!this->interfaces.insert(interface).second) {
        throw std::invalid_argument("Interface conflicts with an interface "
                                    "already in this set.");
    }
}

/**
 * @fn node_interface_set::const_iterator node_interface_set::begin() const throw ()
 *
 * @brief Returns an iterator to the beginning of the node_interface_set.
 *
 * @return a const_iterator pointing to the first node_interface in the set.
 */

/**
 * @fn node_interface_set::const_iterator node_interface_set::end() const throw ()
 *
 * @brief Returns an iterator to the end of the node_interface_set.
 *
 * @return a const_iterator pointing to one increment past the last
 *       node_interface in the set.
 */

namespace {

    /**
     * @internal
     */
    struct interface_id_matches_ :
            std::unary_function<openvrml::node_interface, bool> {
        explicit interface_id_matches_(const std::string & interface_id):
            interface_id(&interface_id)
        {}

        bool operator()(const openvrml::node_interface & interface) const
        {
            static const char eventin_prefix[] = "set_";
            static const char eventout_suffix[] = "_changed";

            return interface.id == *this->interface_id
                || (interface.type == node_interface::exposedfield_id
                    && (eventin_prefix + interface.id == *this->interface_id
                        || (interface.id + eventout_suffix
                            == *this->interface_id)))
                || (interface.type == node_interface::eventin_id
                    && interface.id == eventin_prefix + *this->interface_id)
                || (interface.type == node_interface::eventout_id
                    && interface.id == *this->interface_id + eventout_suffix);
        }

    private:
        const std::string * interface_id;
    };
}

/**
 * @brief Find an interface matching @p id.
 *
 * If no interface is found with an interface identifier that is an exact match
 * for @p id, this method will look for @c set_ and @c _changed variants.
 *
 * @param id    the interface id to look for.
 *
 * @return a const_iterator to the interface, or node_interface_set::end if no
 *      interface is found.
 */
node_interface_set::const_iterator
node_interface_set::find(const std::string & id) const throw ()
{
    return std::find_if(this->begin(), this->end(), interface_id_matches_(id));
}


} // namespace openvrml
